#include "Polygon3.h"
#include "Mesh.h"
#include "ViewCellBsp.h" // TODO: erase this
#include "Mesh.h"
#include "AxisAlignedBox3.h"

Polygon3::Polygon3(): mMaterial(NULL), mParent(NULL)
{}

Polygon3::Polygon3(const VertexContainer &vertices): mVertices(vertices), mMaterial(NULL), mParent(NULL)
{}

Polygon3::Polygon3(MeshInstance *parent): mMaterial(NULL), mParent(parent)
{}


Polygon3::Polygon3(Face *face, Mesh *parentMesh)
{	
	VertexIndexContainer::iterator it = face->mVertexIndices.begin();
	for (; it != face->mVertexIndices.end();  ++it)
	{
		mVertices.push_back(parentMesh->mVertices[*it]);
		mMaterial = parentMesh->mMaterial;
	}
}

Plane3 Polygon3::GetSupportingPlane() const
{
	Vector3 v1 = mVertices[0] - mVertices[1];
	Vector3 v2 = mVertices[2] - mVertices[1];
#ifdef _DEBUG
	Debug << "plane spanned by " <<  v1 << ", " << v2  << endl;
#endif
	return Plane3(mVertices[0], mVertices[1], mVertices[2]);
}

Vector3 Polygon3::GetNormal() const 
{
    return Normalize(CrossProd(mVertices[0] - mVertices[1],
							   mVertices[2] - mVertices[1]));
}

void Polygon3::Split(const Plane3 &partition, 
					 Polygon3 &front, 
					 Polygon3 &back, 
					 VertexContainer &splitPts)
{
	Vector3 ptA = mVertices.back();
	
	int sideA = partition.Side(ptA, SIDE_TOLERANCE);
	
	VertexContainer::const_iterator it;
	
	// find line - plane intersections
	for (it = mVertices.begin(); it != mVertices.end(); ++ it)
	{
		Vector3 ptB = *it;
		int sideB = partition.Side(ptB, SIDE_TOLERANCE);
	
		// vertices on different sides => split
	    if (sideB > 0)
		{
			if (sideA < 0)
			{
				//-- plane - line intersection
				Vector3 splitPt = partition.FindIntersection(ptA, ptB);
			
				// test if split point not too close to previous split point
				if ((splitPts.size() == 0) || 
					(SqrDistance(splitPt, splitPts.back()) > SIDE_TOLERANCE_SQRD))
				{
					// add vertex to both polygons
					front.mVertices.push_back(splitPt);
					back.mVertices.push_back(splitPt);
					
					splitPts.push_back(splitPt);
				}
			}
			front.mVertices.push_back(ptB);
		}
		else if (sideB < 0)
		{
			if (sideA > 0)
			{
				//-- plane - line intersection
				Vector3 splitPt = partition.FindIntersection(ptA, ptB);
				// test if split point not too close to other split point
					// test if split point not too close to previous split point
				if ((splitPts.size() == 0) || 
					(SqrDistance(splitPt, splitPts.back()) > SIDE_TOLERANCE_SQRD))
				{
					// add vertex to both polygons
					front.mVertices.push_back(splitPt);
					back.mVertices.push_back(splitPt);

					splitPts.push_back(splitPt);
				}	
			}
			back.mVertices.push_back(ptB);
		}
		else
		{
			// vertex on plane => add vertex to both polygons
			front.mVertices.push_back(ptB);
			back.mVertices.push_back(ptB); 
		}
	
		ptA = ptB;
		sideA = sideB;
	}
}

float Polygon3::GetArea() const
{
	Vector3 v = CrossProd(mVertices.back(), mVertices.front());
    
    for (int i=0; i < mVertices.size() - 1; ++i)
		v += CrossProd(mVertices[i], mVertices[i+1]);
    
    //Debug << "area2: " << 0.5f * fabs(DotProd(GetNormal(), v)) << endl;  

	return 0.5f * fabs(DotProd(GetNormal(), v));
}

int Polygon3::Side(const Plane3 &plane) const
{
	int classification = ClassifyPlane(plane);
	
	if (classification == BACK_SIDE)
		return -1;
	else if (classification == FRONT_SIDE)
		return 1;

	return 0;
}

int Polygon3::ClassifyPlane(const Plane3 &plane) const
{
	VertexContainer::const_iterator it;

	bool onFrontSide = false;
	bool onBackSide = false;
	
	int count = 0;
	// find possible line-plane intersections
	for (it = mVertices.begin(); it != mVertices.end(); ++ it)
	{
		int side = plane.Side(*it, SIDE_TOLERANCE);
		
		if (side > 0)
			onFrontSide = true;
		else if (side < 0)
			onBackSide = true;
		
		//TODO: check if split goes through vertex
		if (onFrontSide && onBackSide) // split 
		{
			return SPLIT;
		}
		// 3 vertices enough to decide coincident
		else if (((++ count) >= 3) && !onFrontSide && !onBackSide) 
			return COINCIDENT; 
	}

	if (onBackSide)
	{
		return BACK_SIDE;
	}
	else if (onFrontSide)
	{
		return FRONT_SIDE;
	}

	return COINCIDENT; // plane and polygon are coincident
}


Vector3
Polygon3::Center() const
{
	int i;
	Vector3 sum = mVertices[0];
	for (i=1; i < mVertices.size(); i++)
		sum += mVertices[i];
	
	return sum/(float)i;
}


void
Polygon3::Scale(const float scale)
{
	int i;
	Vector3 center = Center();
	for (i=0; i < mVertices.size(); i++) {
		mVertices[i] = center + scale*(mVertices[i] - center);
	}
}

bool Polygon3::Valid() const
{
	if (mVertices.size() < 3)
		return false;

#if 1
	// check if area exceeds certain size
	if (AREA_LIMIT > GetArea())
	{
		//Debug << "area too small: " << GetArea() << endl;
		return false;
	}
#else
	Vector3 vtx = mVertices.back();
	VertexContainer::const_iterator it, it_end = mVertices.end();

	for (it = mVertices.begin(); it != it_end; ++it)
	{
		if (!(SqrDistance(vtx, *it) > SIDE_TOLERANCE_SQRD))
		{
			Debug << "Malformed vertices:\n" << *this << endl;
			return false;
		}
		vtx = *it;
	}
#endif	
	return true;
}

void Polygon3::IncludeInBox(const PolygonContainer &polys, AxisAlignedBox3 &box)
{
	PolygonContainer::const_iterator it, it_end = polys.end();

	for (it = polys.begin(); it != it_end; ++ it)
		box.Include(*(*it));
}